package org.openxma.dsl.platform.dao.impl;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.HashMap;
import java.util.Map;

import org.openxma.dsl.platform.assembler.DtoAssembler;
import org.openxma.dsl.platform.assembler.DtoAssemblerFactory;
import org.openxma.dsl.platform.assembler.DtoAssemblerRegistry;
import org.openxma.dsl.platform.dao.EntityFactory;
import org.openxma.dsl.platform.dao.EntityFactoryRegistry;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.config.BeanPostProcessor;
import org.springframework.stereotype.Component;

@Component
public class MapEntityFactoryRegistry implements EntityFactoryRegistry, DtoAssemblerRegistry, BeanPostProcessor {
    private Map<Class<?>, EntityFactory<?>> entityFactories = new HashMap<Class<?>, EntityFactory<?>>();

    /**
     * this map contains registered {@link DtoAssembler}s with <b>default</b>
     * behavior and associates each to one dto. The key of the map is the dto
     * class, the value is the {@link DtoAssembler}
     */
    private Map<Class<?>, DtoAssembler<?>> dtoAssemblers = new HashMap<Class<?>, DtoAssembler<?>>();

    @SuppressWarnings("unchecked")
    public <T> EntityFactory<T> getEntityFactory(Class<T> entityClass) {
        return (EntityFactory<T>) entityFactories.get(entityClass);
    }

    @SuppressWarnings("unchecked")
    public <T> DtoAssembler<T> getDtoAssembler(Class<T> dtoClass) {
        return (DtoAssembler<T>) dtoAssemblers.get(dtoClass);
    }

    public Object postProcessAfterInitialization(Object bean, String beanName) throws BeansException {
        if (bean instanceof EntityFactory<?>) {
            registerEntityFactoryForClass(bean, bean.getClass());
        } else if (bean instanceof DtoAssemblerFactory) {
            registerDefaultDtoAssemblersForDtoClasses((DtoAssemblerFactory) bean);
        }
        return bean;
    }

    private void registerDefaultDtoAssemblersForDtoClasses(DtoAssemblerFactory dtoAssemblerFactory) {
        for (DtoAssembler<?> dtoAssembler : dtoAssemblerFactory.getDtoAssemblers())
            dtoAssemblers.put(getSupportedDtoClassFromDtoAssemblerClass(dtoAssembler.getClass()), dtoAssembler);
    }

    private Class<?> getSupportedDtoClassFromDtoAssemblerClass(Class<?> beanClass) {
        for (Type type : beanClass.getGenericInterfaces()) {
            if (type instanceof ParameterizedType) {
                ParameterizedType parameterizedType = (ParameterizedType) type;
                if (DtoAssembler.class.equals(parameterizedType.getRawType())) {
                    return (Class<?>) parameterizedType.getActualTypeArguments()[0];
                }
            }
        }

        throw new IllegalArgumentException("missing class of raw type DtoAssembler ");
    }

    private void registerEntityFactoryForClass(Object bean, Class<?> beanClass) {

        for (Type type : beanClass.getGenericInterfaces()) {
            if (type instanceof ParameterizedType) {
                ParameterizedType parameterizedType = (ParameterizedType) type;
                if (EntityFactory.class.equals(parameterizedType.getRawType())) {
                    EntityFactory<?> entityFactory = (EntityFactory<?>) bean;
                    Type[] actualTypeArguments = parameterizedType.getActualTypeArguments();
                    this.entityFactories.put((Class<?>) actualTypeArguments[0], entityFactory);
                }
            } else if (type instanceof Class<?>) {
                registerEntityFactoryForClass(bean, (Class<?>) type);
            }

        }
    }

    public Object postProcessBeforeInitialization(Object bean, String beanName) throws BeansException {
        return bean;
    }

}
