package org.openxma.dsl.platform.hibernate.hql.impl;

import java.io.StringReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.hibernate.HibernateException;
import org.hibernate.MappingException;
import org.hibernate.QueryException;
import org.hibernate.engine.SessionFactoryImplementor;
import org.hibernate.metadata.ClassMetadata;
import org.hibernate.type.CollectionType;
import org.hibernate.type.EntityType;
import org.hibernate.type.Type;
import org.openxma.dsl.platform.hibernate.hql.RewritingHqlParserGen;

import antlr.TokenStreamException;
import antlr.TokenStreamRewriteEngine;

public class RewritingHqlParser extends RewritingHqlParserGen {
    private static final Log log = LogFactory.getLog(RewritingHqlParser.class);
    public static final String ALIAS_REPLACEMENT = "{alias}";
    public static final String ALIAS_REPLACEMENT_ESCAPED = "\\{alias\\}";
    private static final Pattern ALIAS_PATTERN = Pattern.compile("(" + ALIAS_REPLACEMENT_ESCAPED + ".\\w+)");
    private SessionFactoryImplementor sessionFactoryImplementor;

    private RewritingHqlParser(TokenStreamRewriteEngine lexer,SessionFactoryImplementor sessionFactoryImplementor) {
        super(lexer);
        this.sessionFactoryImplementor=sessionFactoryImplementor;
    }

    public static RewritingHqlParser getInstance(String hqlString, SessionFactoryImplementor sessionFactoryImplementor) {
        RewritingHqlLexer lexer = new RewritingHqlLexer(new StringReader(hqlString));
        TokenStreamRewriteEngine rewriteEngine = new TokenStreamRewriteEngine(lexer);
        rewriteEngine.discard(RewritingHqlLexer.WS);
        return new RewritingHqlParser(rewriteEngine,sessionFactoryImplementor);
    }
    
    public String rewriteHqlQuery()  {
        Map<String, List<String>> propertyToAliasMap = getPropertyToAliasMap();
        StringBuffer hqlBuilder = new StringBuffer();
        Matcher matcher = ALIAS_PATTERN.matcher(getRewrittenHql());
        while (matcher.find()) {
            String aliasPath = matcher.group(1);
            String[] split = aliasPath.split("\\.");
            String propertyName = split[1];
            if (getPropertyAliasMap().containsKey(propertyName)) {
            	matcher.appendReplacement(hqlBuilder, getPropertyAliasMap().get(propertyName));
            } else if (getSelectedPropertiesPathMap().containsKey(propertyName)) {
            	List<String> identifierList = getSelectedPropertiesPathMap().get(propertyName);
            	assertPropertyName(propertyName, identifierList);
            	matcher.appendReplacement(hqlBuilder, getSelectedPropertiesPathMap().get(propertyName).get(0));
            } else {
            	List<String> aliasList = propertyToAliasMap.get(propertyName);
            	assertPropertyName(propertyName, aliasList);
            	matcher.appendReplacement(hqlBuilder, aliasPath.replaceAll(RewritingHqlParser.ALIAS_REPLACEMENT_ESCAPED, aliasList.get(0)));
            }
        }
        matcher.appendTail(hqlBuilder);
        return hqlBuilder.toString();
    }

	private void assertPropertyName(String propertyName, List<String> matches) {
		if (matches==null || matches.size() > 1) {
			throw new QueryException("Ambiguous or unassignable property aliases for property '" + propertyName + "' in hql '"
					+ getRewrittenHql().trim() + "' matching query aliases '" + matches + "'");
		}
	}

    
    protected Map<String, List<String>> getPropertyToAliasMap() {
        Map<String, List<String>> propertyToAliasMap = new HashMap<String, List<String>>();
        for (String alias : getClassAliasMap().keySet()) {
            String className = getClassAliasMap().get(alias);
            ClassMetadata classMetadata = getClassMetaData(className);
            addPropertyNames(alias, classMetadata, propertyToAliasMap);
        }
        for (String alias : getJoinAliasMap().keySet()) {
            ClassMetadata classMetadata = getJoinedClassMetaData(alias);
            addPropertyNames(alias, classMetadata, propertyToAliasMap);
        }
        return propertyToAliasMap;
    }

    protected void addPropertyNames(String alias, ClassMetadata classMetadata,
            Map<String, List<String>> propertyToAliasMap) {
        if (classMetadata == null) {
            return;
        }
        for (String propertyName : classMetadata.getPropertyNames()) {
            addPropertyAlias(alias, propertyName, propertyToAliasMap);
        }
    }

    protected void addPropertyAlias(String alias, String propertyName, Map<String, List<String>> propertyToAliasMap) {
		List<String> aliasList = propertyToAliasMap.get(propertyName);
		if (aliasList == null) {
		    aliasList = new ArrayList<String>();
		    propertyToAliasMap.put(propertyName, aliasList);
		}
		aliasList.add(alias);
	}

    protected ClassMetadata getJoinedClassMetaData(String joinAlias) throws HibernateException {
        String joinPath = getJoinAliasMap().get(joinAlias);
        String[] segments = joinPath.split("\\.");
        String joinOrClassAlias = segments[0];
        String joinedPropertyName = segments[1];
        if (getClassAliasMap().containsKey(segments[0])) {
            return getClassMetaData(joinOrClassAlias, joinedPropertyName);
        } else {
            List<String> propertyPathList = getPropertyPathSegments(joinOrClassAlias, joinedPropertyName);
            Iterator<String> iterator = propertyPathList.iterator();
            ClassMetadata classMetaData = getClassMetaData(iterator.next());
            for (; iterator.hasNext(); classMetaData = getClassMetaData(classMetaData.getEntityName(), iterator.next()) ) {
            }
            return classMetaData;
        }
    }

    protected List<String> getPropertyPathSegments(String joinOrClassAlias, String joinedPropertyName) {
        List<String> propertyPathList = new ArrayList<String>(Arrays.asList(joinedPropertyName));
        String alias = joinOrClassAlias;
        while (getJoinAliasMap().containsKey(alias)) {
            String joinPath = getJoinAliasMap().get(alias);
            String[] segments = joinPath.split("\\.");
            alias = segments[0];
            propertyPathList.add(0, segments[1]);
        }
        propertyPathList.add(0, alias);
        return propertyPathList;
    }

    protected ClassMetadata getClassMetaData(String classNameOrAlias, String propertyName) throws HibernateException,
            MappingException {
        ClassMetadata classMetadata = getClassMetaData(classNameOrAlias);
        Type propertyType = classMetadata.getPropertyType(propertyName);
        if (propertyType instanceof EntityType) {
            return getClassMetaData(((EntityType) propertyType).getAssociatedEntityName());
        } else if (propertyType instanceof CollectionType) {
            return getClassMetaData(((CollectionType) propertyType)
                    .getAssociatedEntityName(this.sessionFactoryImplementor));
        } else {
            throw new IllegalStateException("unexpected hibernate type");
        }
    }

    protected ClassMetadata getClassMetaData(String classNameOrAlias) throws HibernateException {
        String importedClassName = getEntityName(classNameOrAlias);
        return this.sessionFactoryImplementor.getClassMetadata(importedClassName);
    }

    protected String getEntityName(String classNameOrAlias) {
        String className = getClassAliasMap().containsKey(classNameOrAlias) ? getClassAliasMap().get(classNameOrAlias) : classNameOrAlias;
        return this.sessionFactoryImplementor.getImportedClassName(className);
    }
    

    public void weakKeywords() throws TokenStreamException {
        int t = LA(1);
        switch (t) {
        case ORDER:
        case GROUP:
            if (LA(2) != LITERAL_by) {
                LT(1).setType(IDENT);
                if (log.isDebugEnabled()) {
                    log.debug("weakKeywords() : new LT(1) token - " + LT(1));
                }
            }
            break;
        default:
            if (LA(0) == FROM && t != IDENT && LA(2) == DOT) {
                HqlTokenWithIndex hqlToken = (HqlTokenWithIndex) LT(1);
                if (hqlToken.isPossibleID()) {
                    hqlToken.setType(IDENT);
                    if (log.isDebugEnabled()) {
                        log.debug("weakKeywords() : new LT(1) token - " + LT(1));
                    }
                }
            }
            break;
        }
    }

    public void handleDotIdent() throws TokenStreamException {
        if (LA(1) == DOT && LA(2) != IDENT) {
            HqlTokenWithIndex t = (HqlTokenWithIndex) LT(2);
            if (t.isPossibleID()) {
                LT(2).setType(IDENT);
                if (log.isDebugEnabled()) {
                    log.debug("handleDotIdent() : new LT(2) token - " + LT(1));
                }
            }
        }
    }
    
}
