package org.openxma.dsl.platform.hibernate.type;

import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.Date;
import java.util.Properties;

import org.hibernate.HibernateException;
import org.hibernate.engine.SessionImplementor;
import org.hibernate.type.BasicType;
import org.hibernate.type.TypeResolver;
import org.hibernate.usertype.ParameterizedType;
import org.hibernate.usertype.UserType;
import org.hibernate.util.EqualsHelper;
import org.openxma.dsl.platform.hibernate.dialect.OpenXmaOracleDialect;
import org.openxma.dsl.platform.valueobject.XmlPersistenceObject;
import org.openxma.dsl.platform.xml.XmlParserFactory;
import org.springframework.beans.BeanWrapperImpl;
import org.springframework.jdbc.support.JdbcUtils;
import org.springframework.orm.hibernate3.support.BlobByteArrayType;
import org.springframework.orm.hibernate3.support.BlobStringType;
import org.springframework.util.ClassUtils;

public class OpenXmaHibernateType implements UserType, ParameterizedType {

	protected Class objectClass = null;
    private final BlobByteArrayType blobArrayType = new BlobByteArrayType();
    private final BlobStringType blobStringType = new BlobStringType();

    public OpenXmaHibernateType() {
		// TODO Auto-generated constructor stub
	}

    public OpenXmaHibernateType(Class objectClass) {
    	this.objectClass = objectClass;
    }

	public Object assemble(Serializable cached, Object owner)
			throws HibernateException {
		if(classInstanceOfClass(objectClass,XmlPersistenceObject.class)) {
			return fromXml(cached.toString());
		}
		else {
			return cached;
		}
	}

	public Object deepCopy(Object value) throws HibernateException {
		return value;
	}

	public Serializable disassemble(Object value) throws HibernateException {
		if(classInstanceOfClass(objectClass,XmlPersistenceObject.class)) {
			return toXml((XmlPersistenceObject) value);
		}
		else {
			return (Serializable)value;
		}
	}

	public boolean equals(Object x, Object y) throws HibernateException {
		return EqualsHelper.equals(x, y);
	}

	public int hashCode(Object x) throws HibernateException {
		return x.hashCode();
	}

	public boolean isMutable() {
		return classInstanceOfClass(objectClass,byte[].class) ||
		   	   classInstanceOfClass(objectClass,Timestamp.class) ||
		   	   classInstanceOfClass(objectClass,Date.class);
	}

	public Object nullSafeGet(ResultSet rs, String[] names, Object owner)
			throws HibernateException, SQLException {
		if(classInstanceOfClass(objectClass,byte[].class)) {
			return (byte[]) blobArrayType.nullSafeGet(rs, names, owner);
		}

		if(classInstanceOfClass(objectClass,XmlPersistenceObject.class)) {
			if(OpenXmaOracleDialect.isXmlTypeRegistered()) {
				Object xmlType = rs.getObject(names[0]);
				BeanWrapperImpl beanWrapperImpl = new BeanWrapperImpl(xmlType);
				return (xmlType != null) ? fromXml((String) beanWrapperImpl
						.getPropertyValue("stringVal")) : null;
			}
			return blobStringType.nullSafeGet(rs, names, owner);
		}

	    return JdbcUtils.getResultSetValue(rs, rs.findColumn(names[0]) , objectClass);
	}

	public void nullSafeSet(PreparedStatement st, Object value, int index)
			throws HibernateException, SQLException {

		if(classInstanceOfClass(objectClass,byte[].class)) {
			blobArrayType.nullSafeSet(st, value, index);
		}
		else if(classInstanceOfClass(objectClass,XmlPersistenceObject.class)) {
			String xml = value.toString();
			if(value instanceof XmlPersistenceObject) {
				xml = toXml((XmlPersistenceObject) value);
			}
			if(OpenXmaOracleDialect.isXmlTypeRegistered()) {
				try {
					Constructor constructor = OpenXmaOracleDialect.loadXmlTypeClass().
							getConstructor(new Class[] { java.sql.Connection.class,
									java.lang.String.class });
					Object xmlType = constructor.newInstance(new Object[] {
							st.getConnection(), xml });
					st.setObject(index, xmlType);
				} catch (Exception e) {
					throw new SQLException(
							"Could not marshall Component object to String for storage");
				}
			}
			else {
				blobStringType.nullSafeSet(st, xml, index);
			}
		}
		else {
			if (null == value) {
	            st.setNull(index, sqlTypes()[0]);
	        }
			else {
	            switch (sqlTypes()[0]) {
	                case Types.DATE:
	                    st.setObject(index, new java.sql.Date(((java.util.Date)value).getTime()), sqlTypes()[0]);
	                    break;
	                case Types.TIMESTAMP:
	                    st.setObject(index, new java.sql.Timestamp(((java.util.Date)value).getTime()), sqlTypes()[0]);
	                    break;
	                case Types.TIME:
	                    st.setObject(index, new java.sql.Time(((java.util.Date)value).getTime()), sqlTypes()[0]);
	                    break;
	                default:
	                    st.setObject(index, value, sqlTypes()[0]);
	            }
	        }
		}

	}

	public Object replace(Object original, Object target, Object owner)
			throws HibernateException {
		return original;
	}

	public Class returnedClass() {
		return objectClass;
	}

	public int[] sqlTypes() {
		if(classInstanceOfClass(objectClass,byte[].class)) {
			return blobArrayType.sqlTypes();
		}
		else if(classInstanceOfClass(objectClass,XmlPersistenceObject.class)) {
			if(OpenXmaOracleDialect.isXmlTypeRegistered()) {
				return new int[] { OpenXmaOracleDialect.getXmlTypeSqlCode() };
			}
			else {
				return blobStringType.sqlTypes();
			}
		}
		else {
		    // TODO test this block
		    TypeResolver typeResolver = new TypeResolver();
		    BasicType basicType = typeResolver.basic(objectClass.getName());		 
		    return basicType.sqlTypes(null);
		}
	}

	public void setParameterValues(Properties parameters) {
    	String objectClassName = parameters.getProperty("objectClass");
		try {
			this.objectClass = ClassUtils.getDefaultClassLoader().loadClass(
					objectClassName);
		} catch (ClassNotFoundException e) {
			throw new HibernateException("Objectclass '"
					+ objectClassName + "' not found");
		}
	}

	protected boolean classInstanceOfClass(Class clazz1, Class clazz2) {
		if(clazz1 == null || clazz2 == null) {
			return false;
		}
		if(clazz1.equals(clazz2)) {
			return true;
		}
		for(int i=0; i < clazz1.getInterfaces().length; i++) {
			if(clazz2.equals(clazz1.getInterfaces()[i])) {
				return true;
			}
		}
		if(clazz1.getSuperclass() != null &&
		   clazz1.getSuperclass() instanceof Class) {
			return classInstanceOfClass((Class) clazz1.getSuperclass(),clazz2);
		}
		return false;
	}

	protected String toXml(XmlPersistenceObject value) {
		return XmlParserFactory.getDefaultXmlParser().toXml(value);
	}

	protected XmlPersistenceObject fromXml(String objectAsString) {
		return (XmlPersistenceObject) XmlParserFactory.getDefaultXmlParser().fromXml(objectAsString);
	}

    public Object nullSafeGet(ResultSet arg0, String[] arg1, SessionImplementor arg2, Object arg3)
            throws HibernateException, SQLException {
        // TODO Auto-generated method stub
        return null;
    }

    public void nullSafeSet(PreparedStatement arg0, Object arg1, int arg2, SessionImplementor arg3)
            throws HibernateException, SQLException {
        // TODO Auto-generated method stub
        
    }
}
