/*******************************************************************************
 * Copyright (c) 2010 s IT Solutions AT Spardat GmbH .
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/
package org.openxma.dsl.platform.query;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * 
 * 
 * <b>Example</b>:
 * 
 * <pre>
 * {
 *     &#064;code
 *     QueryObject customerQuery = new QueryObject();
 *     customerQuery.setFirstResult(1);
 *     customerQuery.setMaxResults(50);
 *     
 *     Expression firstNameEq = customerQuery.get(&quot;firstName&quot;).eq(&quot;foo&quot;);
 *     Expression lastNameLike = customerQuery.get(&quot;lastName&quot;).like(&quot;bar&quot;);
 * 
 *     customerQuery = customerQuery.where(firstNameEq.and(lastNameLike)).orderBy(customerQuery.get(&quot;firstName&quot;).asc());
 * }
 * </pre>
 */
public class QueryObject extends Expression {
    private static final long serialVersionUID = 1L;
    private String queryName;
    private int firstResult;
    private int maxResults;
    private Expression whereExpression;
    private List<Expression> sortOrderExpressions;
    private Map<String, Object> queryArguments;

    public QueryObject() {
    }

    public QueryObject(String queryName) {
        this.queryName = queryName;
    }

    public static QueryObject createNamedQuery(String queryName) {
        return new QueryObject(queryName);
    }

    public static QueryObject createNamedQuery(String queryName, Expression expression) {
        QueryObject result = new QueryObject(queryName);
        if (expression instanceof QueryObject) {
            QueryObject queryObject = (QueryObject) expression;
            result.firstResult = queryObject.firstResult;
            result.maxResults = queryObject.maxResults;
            result.whereExpression = queryObject.whereExpression;
            result.sortOrderExpressions = queryObject.sortOrderExpressions;
        }
        return result;
    }

    public List<Expression> getSortOrderExpressions() {
        if (sortOrderExpressions == null) {
            sortOrderExpressions = new ArrayList<Expression>();
        }
        return sortOrderExpressions;
    }

    public String getQueryName() {
        return queryName;
    }

    public void setQueryName(String queryName) {
        this.queryName = queryName;
    }

    public Map<String, Object> getArguments() {
        if (queryArguments == null) {
            this.queryArguments = new HashMap<String, Object>();
        }
        return queryArguments;
    }

    public Expression getWhereExpression() {
        return whereExpression;
    }

    public int getFirstResult() {
        return firstResult;
    }

    public QueryObject setFirstResult(int firstResult) {
        this.firstResult = firstResult;
        return this;
    }

    public int getMaxResults() {
        return maxResults;
    }

    public QueryObject setMaxResults(int maxResults) {
        this.maxResults = maxResults;
        return this;
    }

    public QueryObject where(Expression expression) {
        this.whereExpression = expression;
        return this;
    }

    public QueryObject orderBy(Expression expression) {
        getSortOrderExpressions().add(expression);
        return this;
    }

    public QueryObject withArgs(String name, Object value) {
        getArguments().put(name, value);
        return this;
    }

    @Override
    public Object evaluate(QueryContext queryContext) {
        return queryContext.accept(this);
    }

}
